import datetime
from mikan.compound import Compound
from mikan.number import Number
from mikan.counter import Counter, MonthCounter, MonthDayCounter

__all__ = ['Date']

class Date(Compound):

    def __init__(self, date: datetime.date) -> None:
        words = [
            Number(date.year) + Counter('年', 'ねん'),
            Number(date.month) + MonthCounter(),
            Number(date.day) + MonthDayCounter(),
        ]
        super().__init__(words)



from .adjective import *
from .compound import *
from .counter import *
from .date import *
from .number import *
from .reading import *
from .utils import *
from .verb import *
from .word import *
from .writing import *

__version__ = '0.1.0'



from typing import List, Type

__all__ = ['Writing']

class Writing(str):

    subclasses: List[Type['Writing']] = []

    def __init_subclass__(cls) -> None:
        cls.subclasses.append(cls)

    @classmethod
    def create(cls, something: object) -> 'Writing':
        string = str(something)
        for subclass in cls.subclasses:
            try:
                writing = subclass(string)
                return writing
            except ValueError:
                pass
        return Writing(string)

    def __add__(self, other: object) -> 'Writing':
        return Writing.create(str(self) + str(other))



from mikan.writing import Writing
import mikan.utils as utils

__all__ = ['Reading']

class Reading(Writing):

    def __new__(cls, string: str) -> 'Reading':
        if not utils.is_kana(string):
            raise ValueError
        return super().__new__(cls, string)



from typing import Dict, Tuple, Any, Union, List
from mikan.base import BaseWord
from mikan.word import Word
from mikan.writing import Writing

__all__ = ['Number']

class Number(Word):
    DIGITS: Dict[int, Dict[str, str]] = {
        1: {'kana': 'いち', 'kanji': '一'},
        2: {'kana': 'に', 'kanji': '二'},
        3: {'kana': 'さん', 'kanji': '三'},
        4: {'kana': 'よん', 'kanji': '四'},
        5: {'kana': 'ご', 'kanji': '五'},
        6: {'kana': 'ろく', 'kanji': '六'},
        7: {'kana': 'なな', 'kanji': '七'},
        8: {'kana': 'はち', 'kanji': '八'},
        9: {'kana': 'きゅう', 'kanji': '九'},
    }

    POS: Dict[int, Dict[str, Any]] = {
        1: {'kana': 'じゅう', 'kanji': '十'},
        2: {'kana': 'ひゃく', 'kanji': '百', 'exceptions': {3: 'さんびゃく', 6: 'ろっぴゃく', 8: 'はっぴゃく'}},
        3: {'kana': 'せん', 'kanji': '千', 'exceptions': {3: 'さんぜん', 8: 'はっせん'}},
    }

    def __init__(
        self,
        *args: Union[int, str, Writing, BaseWord]
    ) -> None:

        digits = None
        for wri in args:
            if isinstance(wri, int):
                digits = wri
            else:
                try:
                    digits = self._from_kanji(str(wri))
                except ValueError:
                    pass

        if digits is None:
            raise ValueError('Not a number')

        self._digits = digits
        writings = self._from_digits(digits)
        super().__init__(*writings)

    @staticmethod
    def _from_digits(num: int, pos: int=0) -> Tuple[str, str]:

        if pos != 0 and pos not in Number.POS:
            raise IndexError
        left = num // 10
        right = num % 10
        lkana = ''
        rkana = ''
        lkanji = ''
        rkanji = ''
        if pos == 0:
            if right > 0:
                rkana = Number.DIGITS[right]['kana']
                rkanji = Number.DIGITS[right]['kanji']
        elif right in Number.POS[pos].get('exceptions', {}):
            rkana = Number.POS[pos]['exceptions'][right]
            rkanji = Number.DIGITS[right]['kanji'] + Number.POS[pos]['kanji']
        elif right > 1:
            rkana = Number.DIGITS[right]['kana'] + Number.POS[pos]['kana']
            rkanji = Number.DIGITS[right]['kanji'] + Number.POS[pos]['kanji']
        elif right == 1:
            rkana = Number.POS[pos]['kana']
            rkanji = Number.POS[pos]['kanji']
        if left > 0:
            lkana, lkanji = Number._from_digits(left, pos + 1)
        return lkana + rkana, lkanji + rkanji

    @staticmethod
    def _from_kanji(kanji: str) -> int:

        if not kanji:
            raise ValueError('Cannot decode kanji into number')

        revdigits = {word['kanji']: digit for digit, word in Number.DIGITS.items()}
        revpos = {word['kanji']: digit for digit, word in Number.POS.items()}

        number = 0
        state = 2
        pos = 0
        remaining = kanji
        while remaining:
            remaining, current = remaining[:-1], remaining[-1]
            if state == 1:
                if current in revpos:
                    pos = revpos[current]
                    state = 2
                else:
                    raise ValueError('Cannot decode kanji into number')
            elif state == 2:
                if current in revdigits:
                    number += revdigits[current] * 10 ** pos
                    state = 1
                elif current in revpos:
                    if pos:
                        number += 10 ** pos
                    pos = revpos[current]
                    state = 2
                else:
                    raise ValueError('Cannot decode kanji into number')

        if state == 2:
            number += 10 ** pos

        return number

    def __int__(self) -> int:
        return self._digits

    @property
    def writings(self) -> List[Writing]:
        return [Writing.create(self._digits)] + super().writings



import abc
from typing import List, Union, cast

from mikan.reading import Reading
from mikan.writing import Writing

class BaseWord(abc.ABC):

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseWord):
            return False
        word: BaseWord = other
        return set(self.writings) == set(word.writings)

    @property
    @abc.abstractmethod
    def writings(self) -> List[Writing]:
        pass

    @property
    def readings(self) -> List[Reading]:
        return cast(List[Reading], list(filter(lambda x: isinstance(x, Reading), self.writings)))

    @abc.abstractmethod
    def __add__(self, other: Union[str, 'BaseWord']) -> 'BaseWord':
        pass



from enum import Enum
from typing import Dict, Callable, Union, List
from mikan.base import BaseWord
from mikan.word import Word
from mikan.writing import Writing

__all__ = ['AdjectiveForm', 'IAdjective', 'YoiAdjective']

AdjectiveForm = Enum('AdjectiveForm', (
    'PRESENT',
    'PAST',
    'TE_FORM',
    'ADVERB',
))

class IAdjective(Word):

    _ENDINGS: Dict[AdjectiveForm, Dict[bool, Callable[[BaseWord], BaseWord]]] = {
        AdjectiveForm.PRESENT: {
            False: lambda x: x + 'い',
            True: lambda x: x + 'くない',
        },
        AdjectiveForm.PAST: {
            False: lambda x: x + 'かった',
            True: lambda x: x + 'くなかった',
        },
        AdjectiveForm.TE_FORM: {
            False: lambda x: x + 'くて',
        },
        AdjectiveForm.ADVERB: {
            False: lambda x: x + 'く',
        },
    }

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        super().__init__(*args)

        base, ending = self.split_okurigana(1)

        if ending != 'い':
            raise ValueError('This is not an i adjective')

        self._stem = base

    def conjugate(
        self,
        forms: Union[AdjectiveForm, List[AdjectiveForm]],
        negative: bool=False
    ) -> BaseWord:

        if isinstance(forms, AdjectiveForm):
            current = forms
        else:
            current = forms[0]

        return self._ENDINGS[current][negative](self._stem)

class YoiAdjective(IAdjective):

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        super().__init__(*args)

        base, ending = self.split_okurigana(2)

        if not ending in ('良い', 'よい'):
            raise ValueError('This is not a yoi adjective')

        self._ii = base + Word('良い', 'いい')

    def conjugate(
        self,
        forms: Union[AdjectiveForm, List[AdjectiveForm]],
        negative: bool=False
    ) -> BaseWord:

        if isinstance(forms, AdjectiveForm):
            current = forms
        else:
            current = forms[0]

        if (current == AdjectiveForm.PRESENT) and not negative:
            return self._ii

        return super().conjugate(current, negative)

    @property
    def writings(self) -> List[Writing]:

        return super().writings + self._ii.writings



from typing import Tuple, List, Union

from mikan.base import BaseWord
from mikan.writing import Writing

import mikan.compound

__all__ = ['Word']

class Word(BaseWord):

    def __init__(
        self,
        *args: Union[str, Writing, BaseWord]
    ) -> None:

        super().__init__()

        self._writings: List[Writing] = []
        for wri in args:
            if isinstance(wri, BaseWord):
                self._writings.extend(wri.writings)
            else:
                self._writings.append(Writing.create(wri))

        if len(self._writings) == 0:
            raise ValueError('No writing provided')

    def __add__(self, other: Union[str, BaseWord]) -> BaseWord:
        if isinstance(other, BaseWord):
            return mikan.compound.Compound((self, other))
        writings = [writing + other for writing in self._writings]
        return Word(*writings)

    @property
    def writings(self) -> List[Writing]:
        return self._writings

    def split_okurigana(self, count: int) -> Tuple['Word', str]:
        okurigana = None
        writings = []

        for writing in self._writings:
            stem, oku = writing[:-count], writing[-count:]
            if okurigana is None:
                okurigana = oku
            elif okurigana != oku:
                raise ValueError('Not okurigana')
            writings.append(Writing.create(stem))

        if okurigana is None:
            # this should not be possible, but...
            raise ValueError('Not okurigana')

        return Word(*writings), okurigana

    def __str__(self) -> str:
        return str(self.writings[0])



from typing import List, Type, Union, Callable, Sequence, Optional
from mikan.base import BaseWord
from mikan.writing import Writing
import mikan.word

__all__ = ['Compound']

class CompoundBase:

    subclasses: List[Type['CompoundBase']] = []

    def __init_subclass__(cls) -> None:
        cls.subclasses.append(cls)

    def __init__(
        self,
        words: Sequence[BaseWord],
        writings: Optional[Sequence[Union[str, Writing]]]=None
    ) -> None:

        super().__init__()

        self._words = list(words)
        self._writings = None
        if writings is not None:
            self._writings = [Writing.create(writing) for writing in writings]

    @property
    def writings(self) -> List[Writing]:
        if self._writings is not None:
            return self._writings

        writings = [Writing.create('')]
        for word in self._words:
            newwritings = []
            for wwriting in word.writings:
                for writing in writings:
                    newwritings.append(writing + wwriting)
            writings = newwritings
        return writings

    def __str__(self) -> str:
        return ''.join([str(word) for word in self._words])

    @staticmethod
    def _combine_writings(
        writings1: List[Writing],
        writings2: List[Writing],
        testfunc: Callable[[Writing], bool]
    ) -> List[Writing]:

        writings = []
        for writing1 in writings1:
            for writing2 in writings2:
                if testfunc(writing1) and testfunc(writing2):
                    writings.append(writing1 + writing2)
        return writings

class Compound(BaseWord):

    def __init__(
        self,
        words: Sequence[BaseWord],
        writings: Optional[Sequence[Union[str, Writing]]]=None
    ) -> None:

        super().__init__()

        compound = None
        for cls in CompoundBase.subclasses:
            try:
                compound = cls(words, writings=writings)
            except ValueError:
                pass

        if compound is None:
            compound = CompoundBase(words, writings=writings)

        self._compound = compound

    @property
    def writings(self) -> List[Writing]:
        return self._compound.writings

    def __str__(self) -> str:
        return str(self._compound)

    def __add__(self, other: Union[str, BaseWord]) -> BaseWord:
        word = other if isinstance(other, BaseWord) else mikan.word.Word(other)
        return Compound((self, word))



from typing import Optional, Tuple, Dict, List, Callable, Iterable
from mikan.exceptions import ConversionError

__all__ = [
    'to_hiragana',
    'to_katakana',
    'to_romaji',
]

def _is_one_hiragana(char: str) -> bool:
    if (ord(char) < 0x3041) or (ord(char) > 0x309E):
        return False
    return True

def _is_one_katakana(char: str) -> bool:
    if (ord(char) < 0x30A0) or (ord(char) > 0x30FE):
        return False
    return True

def _is_one_kana(char: str) -> bool:
    return _is_one_hiragana(char) or _is_one_katakana(char)

def is_hiragana(string: str) -> bool:
    return all(_is_one_hiragana(char) for char in string)

def is_katakana(string: str) -> bool:
    return all(_is_one_katakana(char) for char in string)

def is_kana(string: str) -> bool:
    return all(_is_one_kana(char) for char in string)

_SYMBOLS = [
    ['a',       'あ',       'ア'],
    ['ba',      'ば',       'バ'],
    ['be',      'べ',       'ベ'],
    ['bi',      'び',       'ビ'],
    ['bo',      'ぼ',       'ボ'],
    ['bu',      'ぶ',       'ブ'],
    ['bya',     'びゃ',     'ビャ'],
    ['byo',     'びょ',     'ビョ'],
    ['byu',     'びゅ',     'ビュ'],
    ['cha',     'ちゃ',     'チャ'],
    ['che',     'ちぇ',     'チェ'],
    ['chi',     'ち',       'チ'],
    ['cho',     'ちょ',     'チョ'],
    ['chu',     'ちゅ',     'チュ'],
    ['da',      'だ',       'ダ'],
    ['de',      'で',       'デ'],
    ['dhi',     'でぃ',     'ディ'],
    ['dhu',     'でゅ',     'デュ'],
    ['dji',     'ぢ',       'ヂ'],
    ['djo',     'ぢょ',     'ヂョ'],
    ['do',      'ど',       'ド'],
    ['du',      'どぅ',     'ドゥ'],
    ['dzu',     'づ',       'ヅ'],
    ['e',       'え',       'エ'],
    ['fa',      'ふぁ',     'ファ'],
    ['fe',      'ふぇ',     'フェ'],
    ['fi',      'ふぃ',     'フィ'],
    ['fo',      'ふぉ',     'フォ'],
    ['fu',      'ふ',       'フ'],
    ['ga',      'が',       'ガ'],
    ['ge',      'げ',       'ゲ'],
    ['gi',      'ぎ',       'ギ'],
    ['go',      'ご',       'ゴ'],
    ['gu',      'ぐ',       'グ'],
    ['gya',     'ぎゃ',     'ギャ'],
    ['gyo',     'ぎょ',     'ギョ'],
    ['gyu',     'ぎゅ',     'ギュ'],
    ['ha',      'は',       'ハ'],
    ['he',      'へ',       'ヘ'],
    ['hi',      'ひ',       'ヒ'],
    ['ho',      'ほ',       'ホ'],
    ['hya',     'ひゃ',     'ヒャ'],
    ['hyo',     'ひょ',     'ヒョ'],
    ['hyu',     'ひゅ',     'ヒュ'],
    ['i',       'い',       'イ'],
    ['ja',      'じゃ',     'ジャ'],
    ['je',      'じぇ',     'ジェ'],
    ['ji',      'じ',       'ジ'],
    ['jo',      'じょ',     'ジョ'],
    ['ju',      'じゅ',     'ジュ'],
    ['ka',      'か',       'カ'],
    ['ke',      'け',       'ケ'],
    ['ki',      'き',       'キ'],
    ['ko',      'こ',       'コ'],
    ['ku',      'く',       'ク'],
    ['kya',     'きゃ',     'キャ'],
    ['kyo',     'きょ',     'キョ'],
    ['kyu',     'きゅ',     'キュ'],
    ['ma',      'ま',       'マ'],
    ['me',      'め',       'メ'],
    ['mi',      'み',       'ミ'],
    ['mo',      'も',       'モ'],
    ['mu',      'む',       'ム'],
    ['mya',     'みゃ',     'ミャ'],
    ['myo',     'みょ',     'ミョ'],
    ['myu',     'みゅ',     'ミュ'],
    ['n',       'ん',       'ン'],
    ["n'yo",    'んよ',     'ンヨ'],
    ['na',      'な',       'ナ'],
    ['ne',      'ね',       'ネ'],
    ['ni',      'に',       'ニ'],
    ['no',      'の',       'ノ'],
    ['nu',      'ぬ',       'ヌ'],
    ['nya',     'にゃ',     'ニャ'],
    ['nyo',     'にょ',     'ニョ'],
    ['nyu',     'にゅ',     'ニュ'],
    ['o',       'お',       'オ'],
    ['pa',      'ぱ',       'パ'],
    ['pe',      'ぺ',       'ペ'],
    ['pi',      'ぴ',       'ピ'],
    ['po',      'ぽ',       'ポ'],
    ['pya',     'ぴゃ',     'ピャ'],
    ['pyo',     'ぴょ',     'ピョ'],
    ['pu',      'ぷ',       'プ'],
    ['pyu',     'ぴゅ',     'ピュ'],
    ['ra',      'ら',       'ラ'],
    ['re',      'れ',       'レ'],
    ['ri',      'り',       'リ'],
    ['ro',      'ろ',       'ロ'],
    ['ru',      'る',       'ル'],
    ['rya',     'りゃ',     'リャ'],
    ['ryo',     'りょ',     'リョ'],
    ['ryu',     'りゅ',     'リュ'],
    ['sa',      'さ',       'サ'],
    ['se',      'せ',       'セ'],
    ['sha',     'しゃ',     'シャ'],
    ['she',     'しぇ',     'シェ'],
    ['shi',     'し',       'シ'],
    ['sho',     'しょ',     'ショ'],
    ['shu',     'しゅ',     'シュ'],
    ['so',      'そ',       'ソ'],
    ['su',      'す',       'ス'],
    ['ta',      'た',       'タ'],
    ['te',      'て',       'テ'],
    ['thi',     'てぃ',     'ティ'],
    ['thu',     'てゅ',     'テュ'],
    ['to',      'と',       'ト'],
    ['tsu',     'つ',       'ツ'],
    ['tu',      'とぅ',     'トゥ'],
    ['u',       'う',       'ウ'],
    ['va',      'ゔぁ',     'ヴァ'],
    ['ve',      'ゔぇ',     'ヴェ'],
    ['vi',      'ゔぃ',     'ヴィ'],
    ['vo',      'ゔぉ',     'ヴォ'],
    ['vu',      'ゔ',       'ヴ'],
    ['wa',      'わ',       'ワ'],
    ['whe',     'うぇ',     'ウェ'],
    ['whi',     'うぃ',     'ウィ'],
    ['who',     'うぉ',     'ウォ'],
    ['wo',      'を',       'ヲ'],
    ['ya',      'や',       'ヤ'],
    ['yo',      'よ',       'ヨ'],
    ['yu',      'ゆ',       'ユ'],
    ['za',      'ざ',       'ザ'],
    ['ze',      'ぜ',       'ゼ'],
    ['zo',      'ぞ',       'ゾ'],
    ['zu',      'ず',       'ズ'],
]

def _long_vowel(symbols: Iterable[List[str]]) -> Iterable[List[str]]:
    for symbol in symbols:
        romaji, hiragana, katakana = symbol
        yield symbol
        yield [romaji + '-', hiragana + 'ー', katakana + 'ー']

def _tsu(symbols: Iterable[List[str]]) -> Iterable[List[str]]:
    for symbol in symbols:
        romaji, hiragana, katakana = symbol
        yield symbol
        if not romaji[0] in ('a', 'e', 'n', 'i', 'o', 'u'):
            yield [romaji[0] + romaji, 'っ' + hiragana, 'ッ' + katakana]

def _build_enhancer(
    *args: Callable[[Iterable[List[str]]], Iterable[List[str]]]
) -> Callable[[List[List[str]]], Iterable[List[str]]]:

    def _enhance(symbols: Iterable[List[str]]) -> Iterable[List[str]]:
        for func in args:
            symbols = func(symbols)
        return symbols
    return _enhance

class Converter:

    def __init__(self) -> None:

        self._symbols: Dict[str, List[str]] = {}

        enhancer = _build_enhancer(_long_vowel, _tsu)

        for symbol in enhancer(_SYMBOLS):
            romaji, hiragana, katakana = symbol

            if romaji in self._symbols:
                raise Exception(f'duplicate symbol {romaji}')
            self._symbols[romaji] = symbol

            if hiragana in self._symbols:
                raise Exception(f'duplicate symbol {hiragana}')
            self._symbols[hiragana] = symbol

            if katakana in self._symbols:
                raise Exception(f'duplicate symbol {katakana}')
            self._symbols[katakana] = symbol

        self._symbols['/'] = ['/', '・', '・']
        self._symbols['・'] = ['/', '・', '・']

    def next_symbol(self, string: str) -> Tuple[Optional[str], Optional[List[str]]]:

        for length in range(4, 0, -1):
            sub = string[0:length]
            if sub in self._symbols:
                return sub, self._symbols[sub]
        return None, None

    def _convert(self, string: str, index: int) -> str:

        todecode = string

        symbols = []
        while todecode:
            sub, symbol = self.next_symbol(todecode)
            if (sub is None) or (symbol is None):
                raise ConversionError(f"Cannot convert {todecode} in {string}")
            symbols.append(symbol[index])
            todecode = todecode[len(sub):]

        return ''.join(symbols)

    def to_romaji(self, string: str) -> str:

        return self._convert(string, 0)

    def to_hiragana(self, string: str) -> str:

        return self._convert(string, 1)

    def to_katakana(self, string: str) -> str:

        return self._convert(string, 2)

def to_katakana(string: str) -> str:
    converter = Converter()
    return converter.to_katakana(string)

def to_hiragana(string: str) -> str:
    converter = Converter()
    return converter.to_hiragana(string)

def to_romaji(string: str) -> str:
    converter = Converter()
    return converter.to_romaji(string)



from typing import List, Optional, Tuple, Callable, Dict
from mikan.compound import CompoundBase
from mikan.number import Number
from mikan.word import Word
from mikan.reading import Reading
from mikan.writing import Writing

__all__ = [
    'Counter',
    'DayHourCounter',
    'MonthDayCounter',
    'MonthCounter',
    'TsuCounter',
]

class Counter(Word):
    pass

class CounterCompound(CompoundBase):

    __EXCEPTIONS: Dict[str, Dict[str, Callable[[str, str], List[str]]]] = {
        'いち': {
            'ふ': lambda x, y: [x[:-1] + 'っぷ' + y[1:]],
            'ひ': lambda x, y: [x[:-1] + 'っぴ' + y[1:]],
            'ほ': lambda x, y: [x[:-1] + 'っぽ' + y[1:]],
        },
        'さん': {
            'ふ': lambda x, y: [x[:-1] + 'っぷ' + y[1:]],
            'ひ': lambda x, y: [x + 'び' + y[1:]],
            'ほ': lambda x, y: [x + 'ぼ' + y[1:]],
        },
        'ろく': {
            'ふ': lambda x, y: [x[:-1] + 'っぷ' + y[1:]],
            'ひ': lambda x, y: [x[:-1] + 'っぴ' + y[1:]],
            'ほ': lambda x, y: [x[:-1] + 'っぽ' + y[1:]],
        },
        'はち': {
            'ふ': lambda x, y: [x[:-1] + 'っぷ' + y[1:]],
            'ひ': lambda x, y: [x[:-1] + 'っぴ' + y[1:], x + y],
            'ほ': lambda x, y: [x[:-1] + 'っぽ' + y[1:], x + y],
        },
        'じゅう': {
            'ふ': lambda x, y: [x[:-1] + 'っぷ' + y[1:]],
            'ひ': lambda x, y: [x[:-1] + 'っぴ' + y[1:]],
            'ほ': lambda x, y: [x[:-1] + 'っぽ' + y[1:]],
        }
    }

    def _find_func(self, writing1: str, writing2: str) -> Optional[Callable[[str, str], List[str]]]:
        suffix = None
        if writing1[-3:] in self.__EXCEPTIONS:
            suffix = writing1[-3:]
        elif writing1[-2:] in self.__EXCEPTIONS:
            suffix = writing1[-2:]
        if suffix is None:
            return None

        return self.__EXCEPTIONS[suffix].get(writing2[0])

    def __init__(self, words: List[Word], writings: Optional[List[Writing]]=None) -> None:

        if (
            (len(words) != 2) or
            not (isinstance(words[0], Number) and isinstance(words[1], Counter))
        ):
            raise ValueError

        number, counter = words

        if writings is None:
            writings = []

            for wri1 in number.writings:
                for wri2 in counter.writings:
                    if isinstance(wri1, Reading) and isinstance(wri2, Reading):
                        func = self._find_func(wri1, wri2)
                        if func is not None:
                            writings.extend([Reading(reading) for reading in func(wri1, wri2)])
                        else:
                            writings.append(wri1 + wri2)
                    elif (not isinstance(wri1, Reading)) and (not isinstance(wri2, Reading)):
                        writings.append(wri1 + wri2)

        super().__init__(words, writings=writings)

class DayHourCounter(Word):

    def __init__(self) -> None:
        super().__init__('時', 'じ')

class DayHourCounterCompound(CompoundBase):

    __EXCEPTIONS = {
        4: ['よじ'],
        9: ['くじ'],
    }

    def __init__(
        self,
        words: Tuple[Number, Counter],
        writings: Optional[List[Writing]]=None
    ) -> None:

        if (
            (len(words) != 2) or
            not (isinstance(words[0], Number) and isinstance(words[1], DayHourCounter))
        ):
            raise ValueError

        number, counter = words

        if writings is None:
            writings = []
            value = int(number)
            if value in self.__EXCEPTIONS:
                writings = [Reading(reading) for reading in self.__EXCEPTIONS[value]]
            else:
                writings.extend(self._combine_writings(
                    number.writings,
                    counter.writings,
                    lambda x: isinstance(x, Reading)
                ))
            writings.extend(self._combine_writings(
                number.writings,
                counter.writings,
                lambda x: not isinstance(x, Reading)
            ))

        super().__init__(words, writings=writings)

class MonthDayCounter(Word):

    def __init__(self) -> None:
        super().__init__('日', 'にち')

class MonthDayCounterCompound(CompoundBase):

    __EXCEPTIONS = {
        1: ['ついたち'],
        2: ['ふつか'],
        3: ['みっか'],
        4: ['よっか'],
        5: ['いつか'],
        6: ['むいか'],
        7: ['なのか'],
        8: ['ようか'],
        9: ['ここのか'],
        10: ['とおか'],
        14: ['じゅうよっか'],
        20: ['はつか'],
        24: ['にじゅうよっか'],
    }

    def __init__(
        self,
        words: Tuple[Number, Counter],
        writings: Optional[List[Writing]]=None
    ) -> None:

        if (
            (len(words) != 2) or
            not (isinstance(words[0], Number) and isinstance(words[1], MonthDayCounter))
        ):
            raise ValueError

        number, counter = words

        if writings is None:
            writings = []
            value = int(number)
            if value in self.__EXCEPTIONS:
                writings = [Reading(reading) for reading in self.__EXCEPTIONS[value]]
            else:
                writings.extend(self._combine_writings(
                    number.writings,
                    counter.writings,
                    lambda x: isinstance(x, Reading)
                ))
            writings.extend(self._combine_writings(
                number.writings,
                counter.writings,
                lambda x: not isinstance(x, Reading)
            ))

        super().__init__(words, writings=writings)

class MonthCounter(Word):

    def __init__(self) -> None:
        super().__init__('月', 'がつ')

class MonthCounterCompound(CompoundBase):

    __EXCEPTIONS = {
        4: ['しがつ'],
        7: ['しちがつ'],
        9: ['くがつ'],
    }

    def __init__(
        self,
        words: Tuple[Number, Counter],
        writings: Optional[List[Writing]]=None
    ) -> None:

        if (
            (len(words) != 2) or
            not (isinstance(words[0], Number) and isinstance(words[1], MonthCounter))
        ):
            raise ValueError

        number, counter = words

        if writings is None:
            writings = []
            value = int(number)
            if value in self.__EXCEPTIONS:
                writings = [Reading(reading) for reading in self.__EXCEPTIONS[value]]
            else:
                writings.extend(self._combine_writings(
                    number.writings,
                    counter.writings,
                    lambda x: isinstance(x, Reading)
                ))
            writings.extend(self._combine_writings(
                number.writings,
                counter.writings,
                lambda x: not isinstance(x, Reading)
            ))

        super().__init__(words, writings=writings)

class TsuCounter(Word):

    def __init__(self) -> None:
        super().__init__('つ')

class TsuCounterCompound(CompoundBase):

    __EXCEPTIONS = {
        1: ['ひとつ'],
        2: ['ふたつ'],
        3: ['みっつ'],
        4: ['よっつ'],
        5: ['いつつ'],
        6: ['むっつ'],
        7: ['ななつ'],
        8: ['やっつ'],
        9: ['ここのつ'],
    }

    def __init__(
        self,
        words: Tuple[Number, Counter],
        writings: Optional[List[Writing]]=None
    ) -> None:

        if (
            (len(words) != 2) or
            not (isinstance(words[0], Number) and isinstance(words[1], TsuCounter))
        ):
            raise ValueError

        number, counter = words

        if writings is None:
            writings = []
            value = int(number)
            if value in self.__EXCEPTIONS:
                writings = [Reading(reading) for reading in self.__EXCEPTIONS[value]]
            for wri1 in number.writings:
                for wri2 in counter.writings:
                    if not isinstance(wri1, Reading):
                        writings.append(wri1 + wri2)

        super().__init__(words, writings=writings)



from enum import Enum
from typing import Dict, Tuple, Callable, Union, List, Optional

from mikan.adjective import IAdjective
from mikan.base import BaseWord
from mikan.compound import Compound
from mikan.word import Word
from mikan.writing import Writing

__all__ = [
    'VerbForm',
    'GodanVerb',
    'IchidanVerb',
    'KuruVerb',
    'SuruVerb',
    'IkuVerb',
]

VerbForm = Enum('VerbForm', (
    'PRESENT',
    'PAST',
    'IMPERATIVE',
    'TE_FORM',
    'CONDITIONAL_EBA',
    'CONDITIONAL_RA',
    'PRESUMPTIVE',
    'VOLITIONAL',
    'POTENTIAL',
    'PASSIVE',
    'CAUSATIVE',
    'TAI',
))

class Verb:
    """A Verb."""

    _ENDINGS: Dict[
        VerbForm,
        Dict[
            bool,
            Dict[
                bool,
                Callable[
                    [Tuple[Dict[str, BaseWord], bool]],
                    BaseWord
                ]
            ]
        ]
    ] = {
        VerbForm.PRESENT: {
            False: {
                False: lambda x: x[0]['self'],
                True: lambda x: x[0]['masu'] + 'ます',
            },
            True: {
                False: lambda x: x[0]['nai'] + 'ない',
                True: lambda x: x[0]['masu'] + 'ません',
            },
        },
        VerbForm.PAST: {
            False: {
                False: lambda x: x[0]['ta'] + ('だ' if x[1] else 'た'),
                True: lambda x: x[0]['masu'] + 'ました',
            },
            True: {
                False: lambda x: x[0]['nai'] + 'なかった',
                True: lambda x: x[0]['masu'] + 'ませんでした',
            },
        },
        VerbForm.IMPERATIVE: {
            False: {
                False: lambda x: x[0]['imp'],
            },
            True: {
                False: lambda x: x[0]['self'] + 'な',
            },
        },
        VerbForm.TE_FORM: {
            False: {
                False: lambda x: x[0]['ta'] + ('で' if x[1] else 'て'),
            },
            True: {
                False: lambda x: x[0]['nai'] + 'なくて',
            },
        },
        VerbForm.CONDITIONAL_EBA: {
            False: {
                False: lambda x: x[0]['e'] + 'ば',
            },
            True: {
                False: lambda x: x[0]['nai'] + 'なければ',
            },
        },
        VerbForm.CONDITIONAL_RA: {
            False: {
                False: lambda x: x[0]['ta'] + ('だ' if x[1] else 'た') + 'ら',
                True: lambda x: x[0]['masu'] + 'ましたら',
            },
            True: {
                False: lambda x: x[0]['nai'] + 'なかったら',
                True: lambda x: x[0]['masu'] + 'ませんでしたら',
            },
        },
        VerbForm.PRESUMPTIVE: {
            False: {
                False: lambda x: x[0]['self'] + 'だろう',
                True: lambda x: x[0]['self'] + 'でしょう',
            },
            True: {
                False: lambda x: x[0]['nai'] + 'ないだろう',
                True: lambda x: x[0]['nai'] + 'ないでしょう',
            },
        },
        VerbForm.VOLITIONAL: {
            False: {
                False: lambda x: x[0]['vol'] + 'う',
                True: lambda x: x[0]['masu'] + 'ましょう',
            },
        },
    }

    def __init__(self, forms: Tuple[Dict[str, BaseWord], bool]) -> None:

        self._forms = forms

    def conjugate(
        self,
        forms: Optional[Union[VerbForm, List[VerbForm]]]=None,
        polite: bool=False,
        negative: bool=False
    ) -> BaseWord:
        """Conjugate the verb in the given tense."""

        if forms is None:
            forms = VerbForm.PRESENT

        if isinstance(forms, VerbForm):
            current = forms
            follow = [VerbForm.PRESENT]
        else:
            current = forms[0]
            follow = forms[1:] if len(forms) > 1 else [VerbForm.PRESENT]

        if current == VerbForm.POTENTIAL:
            return IchidanVerb(self._forms[0]['pot'] + 'る').conjugate(follow, polite, negative)
        if current == VerbForm.PASSIVE:
            return IchidanVerb(self._forms[0]['pas'] + 'れる').conjugate(follow, polite, negative)
        if current == VerbForm.CAUSATIVE:
            return IchidanVerb(self._forms[0]['cau'] + 'せる').conjugate(follow, polite, negative)
        if current == VerbForm.TAI:
            return IAdjective(self._forms[0]['masu'] + 'たい')

        return self._ENDINGS[current][negative][polite](self._forms)

class IchidanVerb(Verb, Word):
    """An ichidan (AKA group 2) verb."""

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        Word.__init__(self, *args)

        base, ending = self.split_okurigana(1)
        if ending != 'る':
            raise ValueError('This is not an ichidan verb')

        Verb.__init__(self, ({
            'ta': base,
            'pot': base + 'られ',
            'pas': base + 'ら',
            'cau': base + 'さ',
            'nai': base,
            'masu': base,
            'imp': base + 'ろ',
            'e': base + 'れ',
            'vol': base + 'よ',
            'self': self,
        }, False))

class GodanVerb(Verb, Word):
    """A godan (AKA group 1) verb."""

    _GODAN_ENDINGS = {
        'う': ({
            'i': 'い',
            'ta': 'っ',
            'a': 'わ',
            'e': 'え',
            'o': 'お',
        }, False),
        'る': ({
            'i': 'り',
            'ta': 'っ',
            'a': 'ら',
            'e': 'れ',
            'o': 'ろ',
        }, False),
        'む' : ({
            'i': 'み',
            'ta': 'ん',
            'a': 'ま',
            'e': 'め',
            'o': 'も',
        }, True),
        'ぬ' : ({
            'i': 'に',
            'ta': 'ん',
            'a': 'な',
            'e': 'ね',
            'o': 'の',
        }, True),
        'く': ({
            'i': 'き',
            'ta': 'い',
            'a': 'か',
            'e': 'け',
            'o': 'こ',
        }, False),
        'ぐ': ({
            'i': 'ぎ',
            'ta': 'い',
            'a': 'が',
            'e': 'げ',
            'o': 'ご',
        }, True),
        'ぶ': ({
            'i': 'び',
            'ta': 'ん',
            'a': 'ば',
            'e': 'べ',
            'o': 'ぼ',
        }, True),
        'す': ({
            'i': 'し',
            'ta': 'し',
            'a': 'さ',
            'e': 'せ',
            'o': 'そ',
        }, False),
        'つ': ({
            'i': 'ち',
            'ta': 'っ',
            'a': 'た',
            'e': 'て',
            'o': 'と',
        }, False),
    }

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        Word.__init__(self, *args)

        base, ending = self.split_okurigana(1)
        if ending not in self._GODAN_ENDINGS:
            raise ValueError('Not a godan verb')

        forms = ({
            'ta': base + self._GODAN_ENDINGS[ending][0]['ta'],
            'pot': base + self._GODAN_ENDINGS[ending][0]['e'],
            'pas': base + self._GODAN_ENDINGS[ending][0]['a'],
            'cau': base + self._GODAN_ENDINGS[ending][0]['a'],
            'nai': base + self._GODAN_ENDINGS[ending][0]['a'],
            'masu': base + self._GODAN_ENDINGS[ending][0]['i'],
            'imp': base + self._GODAN_ENDINGS[ending][0]['e'],
            'e': base + self._GODAN_ENDINGS[ending][0]['e'],
            'vol': base + self._GODAN_ENDINGS[ending][0]['o'],
            'self': self,
        }, self._GODAN_ENDINGS[ending][1])

        Verb.__init__(self, forms)

class KuruVerb(Verb, Word):
    """The kuru (come) verb."""

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        Word.__init__(self, *args)

        base_writings = []
        for writing in self.writings:
            ending = writing[-2:]
            if not ending in ('来る', 'くる'):
                raise ValueError('Not a kuru verb')
            base_writings.append(writing[:-2])
        self._base = Word(*base_writings)

        Verb.__init__(self, ({
            'ta': Word('来', 'き'),
            'pot': Word('来られ', 'こられ'),
            'pas': Word('来ら', 'こら'),
            'cau': Word('来さ', 'こさ'),
            'nai': Word('来', 'こ'),
            'masu': Word('来', 'き'),
            'imp': Word('来い', 'こい'),
            'e': Word('来れ', 'くれ'),
            'vol': Word('来よ', 'こよ'),
            'self': Word('来る', 'くる'),
        }, False))

    def conjugate(
        self,
        forms: Optional[Union[VerbForm, List[VerbForm]]]=None,
        polite: bool=False,
        negative: bool=False
    ) -> BaseWord:

        conj = super().conjugate(forms, polite, negative)
        return self._base + conj if self._base else conj

class SuruVerb(Verb, Word):
    """The suru (do) verb."""

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        if len(args) > 0:
            self._base: Optional[BaseWord] = Word(*args)
            word: BaseWord = Compound((self._base, Word('する')))
        else:
            self._base = None
            word = Word('する')

        Word.__init__(self, word)
        Verb.__init__(self, ({
            'ta': Word('し'),
            'pot': Word('でき'),
            'pas': Word('さ'),
            'cau': Word('さ'),
            'nai': Word('し'),
            'masu': Word('し'),
            'imp': Word('しろ'),
            'e': Word('すれ'),
            'vol': Word('しよ'),
            'self': self,
        }, False))

    def conjugate(
        self,
        forms: Optional[Union[VerbForm, List[VerbForm]]]=None,
        polite: bool=False,
        negative: bool=False
    ) -> BaseWord:

        conj = super().conjugate(forms, polite, negative)
        return self._base + conj if self._base else conj

class IkuVerb(GodanVerb):

    def __init__(self, *args: Union[str, Writing, BaseWord]) -> None:

        super().__init__(*args)

        for writing in self.writings:
            ending = writing[-2:]
            if not ending in ('行く', 'いく'):
                raise ValueError('Not an iku verb')

        sbase, _ = self.split_okurigana(1)

        self._forms[0]['ta'] = sbase + 'っ'
